/*
 * Harry - A Tool for Measuring String Similarity
 * Copyright (C) 2013-2014 Konrad Rieck (konrad@mlsec.org)
 * --
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.  This program is distributed without any
 * warranty. See the GNU General Public License for more details. 
 */

/**
 * @defgroup measures Measure interface
 *
 * Interface and functions for computing similarity measures for strings
 *
 * @{
 */

#include "config.h"
#include "common.h"
#include "harry.h"
#include "util.h"
#include "input.h"
#include "hstring.h"
#include "measures.h"
#include "vcache.h"

/* Module headers */
#include "dist_bag.h"
#include "dist_compression.h"
#include "dist_damerau.h"
#include "dist_hamming.h"
#include "dist_jarowinkler.h"
#include "dist_kernel.h"
#include "dist_lee.h"
#include "dist_levenshtein.h"
#include "dist_osa.h"
#include "kern_distance.h"
#include "kern_spectrum.h"
#include "kern_subsequence.h"
#include "kern_wdegree.h"
#include "sim_coefficient.h"


/* External variables */
extern config_t cfg;

/* Global variables */
static cfg_int global_cache;
static int idx = 0;

/* Module interfaces */
measure_t func[] = {
    {"dist_bag", dist_bag_config, dist_bag_compare},
    {"dist_compression", dist_compression_config, dist_compression_compare},
    {"dist_ncd", dist_compression_config, dist_compression_compare},
    {"dist_damerau", dist_damerau_config, dist_damerau_compare},
    {"dist_hamming", dist_hamming_config, dist_hamming_compare},
    {"dist_jaro", dist_jaro_config, dist_jaro_compare},
    {"dist_jarowinkler", dist_jarowinkler_config, dist_jarowinkler_compare},
    {"dist_kernel", dist_kernel_config, dist_kernel_compare},
    {"dist_lee", dist_lee_config, dist_lee_compare},
    {"dist_levenshtein", dist_levenshtein_config, dist_levenshtein_compare},
    {"dist_edit", dist_levenshtein_config, dist_levenshtein_compare},
    {"dist_osa", dist_osa_config, dist_osa_compare},
    {"kern_distance", kern_distance_config, kern_distance_compare},
    {"kern_dsk", kern_distance_config, kern_distance_compare},
    {"kern_spectrum", kern_spectrum_config, kern_spectrum_compare},
    {"kern_ngram", kern_spectrum_config, kern_spectrum_compare},
    {"kern_subsequence", kern_subsequence_config, kern_subsequence_compare},
    {"kern_ssk", kern_subsequence_config, kern_subsequence_compare},
    {"kern_wdegree", kern_wdegree_config, kern_wdegree_compare},
    {"kern_wdk", kern_wdegree_config, kern_wdegree_compare},
    {"sim_braun", sim_braun_config, sim_braun_compare},
    {"sim_dice", sim_dice_config, sim_dice_compare},
    {"sim_czekanowski", sim_dice_config, sim_dice_compare},
    {"sim_jaccard", sim_jaccard_config, sim_jaccard_compare},
    {"sim_kulczynski", sim_kulczynski_config, sim_kulczynski_compare},
    {"sim_otsuka", sim_otsuka_config, sim_otsuka_compare},
    {"sim_ochiai", sim_otsuka_config, sim_otsuka_compare},
    {"sim_simpson", sim_simpson_config, sim_simpson_compare},
    {"sim_sokal", sim_sokal_config, sim_sokal_compare},
    {"sim_anderberg", sim_sokal_config, sim_sokal_compare},
    {NULL}
};

/**
 * Find best matching measure name.
 * @param name to check
 */
int measure_match(const char *name)
{
    int r = 0;
    float d, dist = FLT_MAX;

    for (int i = 0; func[i].name; i++) {
        /* Match full name */
        d = fabs(strcasecmp(name, func[i].name));
        if (d < dist) {
            r = i, dist = d;
        }

        /* Match from second token */
        char *token = strchr(func[i].name, '_') + 1;
        d = fabs(strcasecmp(name, token));
        if (d < dist) {
            r = i, dist = d;
        }
    }
    
    if (dist > 0) {
        r = 0;
        warning("Unknown measure '%s'. Using '%s' instead.", name, 
                func[r].name);
    }

    return r;
}

/**
 * Configures the measure for a given similarity measure.
 * @param name Name of similarity measure
 * @return name of selected similarity measure
 */
char *measure_config(const char *name)
{
    const char *cfg_str;

    /* Set delimiters */
    config_lookup_string(&cfg, "measures.token_delim", &cfg_str);
    if (strlen(cfg_str) > 0)
        hstring_delim_set(cfg_str);
    else
        hstring_delim_reset();

    /* Enable global cache */
    config_lookup_int(&cfg, "measures.global_cache", &global_cache);

    /* Configure */
    idx = measure_match(name);
    func[idx].measure_config();
    return func[idx].name;
}

/** 
 * Print list of supported similarity measures
 * @param f File stream 
 */
void measure_fprint(FILE *f)
{
        fprintf(f,
           "    dist_bag             Bag distance\n"
           "    dist_compression     Normalized compression distance (NCD)\n"
           "    dist_damerau         Damerau-Levenshtein distance\n"
           "    dist_hamming         Hamming distance\n"
           "    dist_jaro            Jaro distance\n"
           "    dist_jarowinkler     Jaro-Winkler distance\n"
           "    dist_kernel          Kernel substitution distance\n"
           "    dist_lee             Lee distance\n"
           "    dist_levenshtein     Levenshtein distance\n"
           "    dist_osa             Optimal string alignment (OSA) distance\n"
           "    kern_distance        Distance substitution kernel (DSK)\n"
           "    kern_spectrum        Spectrum kernel\n"
           "    kern_subsequence     Subsequence kernel (SSK)\n"
           "    kern_wdegree         Weighted-degree kernel (WDK)\n"
           "    sim_braun            Braun-Blanquet coefficient\n"
           "    sim_dice             Soerensen-Dice coefficient\n"
           "    sim_jaccard          Jaccard coefficient\n"
           "    sim_kulczynski       second Kulczynski coefficient\n"
           "    sim_otsuka           Otsuka coefficient\n"
           "    sim_simpson          Simpson coefficient\n"
           "    sim_sokal            Sokal-Sneath coefficient\n"
    );

}

/**
 * Compares two strings with the given similarity measure.
 * @param x first string
 * @param y second second
 * @return similarity/dissimilarity value
 */
double measure_compare(hstring_t x, hstring_t y)
{
    if (!global_cache)
        return func[idx].measure_compare(x, y);

    uint64_t xyk = hstring_hash2(x, y);
    float m = 0;

    if (!vcache_load(xyk, &m, ID_COMPARE)) {
        m = func[idx].measure_compare(x, y);
        vcache_store(xyk, m, ID_COMPARE);
    }
    return m;
}

/** @} */
